@@ -1,4 +1,4 @@
-# Copyright (c) 2011-2013 Bertrand Janin <b@janin.com>
+# Copyright (c) 2011-2015 Bertrand Janin <b@janin.com>
 #
 # Permission to use, copy, modify, and/or distribute this software for any
 # purpose with or without fee is hereby granted, provided that the above
@@ -15,6 +15,7 @@
 import csv
 import sys
 import os
+import subprocess
 import requests
 import tempfile
 import webbrowser
@@ -83,6 +84,7 @@ class CartmanApp(object):
         self.add_comment = args.add_comment
         self.message = args.message
         self.template = args.template
+        self.message_file = args.message_file
 
         self.ensure_directories()
         self.read_config()
@@ -101,8 +103,6 @@ class CartmanApp(object):
             self.print_function_help(func_name)
             return
 
-        self.login()
-
         try:
             output = func(*args.parameters)
         except exceptions.InvalidParameter as ex:
@@ -155,10 +155,15 @@ class CartmanApp(object):
         if auth_type not in AUTH_TYPES:
             supported_auths = ", ".join(sorted(AUTH_TYPES.keys()))
             msg = ("invalid auth setting '{}', supported: {}"
-                  .format(auth_type, supported_auths))
+                   .format(auth_type, supported_auths))
             raise exceptions.ConfigError(msg)
         self.auth_type = auth_type
 
+        if cp.has_option(self.site, "editor"):
+            self.config_editor = cp.get(self.site, "editor")
+        else:
+            self.config_editor = os.getenv("EDITOR")
+
         self.required_fields = ["To", "Component", "Subject", "Priority"]
 
     def get_form_token(self):
@@ -181,7 +186,21 @@ class CartmanApp(object):
         JavaScript dictionary exposed on the query page.
 
         """
-        return text.extract_properties(self.get("/query").text)
+        # NOTE: lack of "order" parameter disables "max" request too
+        r = self.get("/query?max=1&order=priority")
+        return text.extract_properties(r.text)
+
+    def get_property_options(self):
+        """Return all the property options, with option groups expanded."""
+        options = {}
+        for name, prop in self.get_properties().items():
+            if "options" not in prop:
+                continue
+            options[name] = prop["options"]
+            if "optgroups" in prop:
+                for optgroup in prop["optgroups"]:
+                    options[name] += optgroup["options"]
+        return options
 
     def check_version(self, raw_html):
         """Print a warning if the version of Trac is unsupported."""
@@ -203,13 +222,17 @@ class CartmanApp(object):
         self.trac_version = trac_version
 
     def editor(self, filename):
-        """Spawn the default editor ($EDITOR env var)."""
+        """Spawn the default editor ($EDITOR env var or editor configuration
+        item)."""
 
-        if not os.getenv("EDITOR"):
-            raise exceptions.FatalError("unable to get an EDITOR environment "
-                                        "variable")
+        if not self.config_editor:
+            raise exceptions.FatalError("no editor configured (EDITOR "
+                                        "environment variable or editor "
+                                        "configuration item)")
 
-        os.system("$EDITOR '{}'".format(filename))
+        if subprocess.call([self.config_editor, filename]) != 0:
+            raise exceptions.FatalError("there was a problem running the "
+                                        "editor")
 
     def input(self, prompt):
         return raw_input(prompt)
@@ -227,6 +250,7 @@ class CartmanApp(object):
                               HTTP return code (default: True).
 
         """
+
         r = self.session.get(self.base_url + query_string, data=data)
 
         if r.status_code >= 400 and handle_errors:
@@ -286,10 +310,6 @@ class CartmanApp(object):
                    .format(r.request.url))
             raise exceptions.LoginError(msg)
 
-        # Load a page to get the new cookies.
-        if r.status_code != 302:
-            r = self.get("/")
-
         self.logged_in = True
 
     def get_dicts(self, query_string):
@@ -447,6 +467,64 @@ class CartmanApp(object):
     # documentation.
     #
 
+    def run_change(self, ticket_id, *values):
+        """Make change to the given ticket_id. This command does not return
+        anything if successful.
+
+        TODO: support spanning an editor to change field values.
+
+        usage: cm change ticket_id field=value [field=value...]
+
+        """
+        ticket_id = text.validate_id(ticket_id)
+
+        if not values:
+            raise exceptions.InvalidParameter("should provide at least one "
+                                              "field change")
+        fields_data = {}
+        for v in values:
+            s = v.split('=', 1)
+            if len(s) != 2:
+                raise exceptions.InvalidParameter(
+                        "invalid value '{}', should be a field=value "
+                        "pair".format(v))
+            field = s[0].strip()
+            value = s[1]
+            fields_data["field_" + field] = value
+
+        self.login()
+
+        # Load the timestamps from the ticket page.
+        r = self.get("/ticket/{}".format(ticket_id))
+        timestamps = self._extract_timestamps(r.text)
+
+        if self.message:
+            comment = self.message
+        elif self.add_comment:
+            comment = self._read_comment()
+        else:
+            comment = ""
+
+        data = {
+            "action": "leave",
+            "comment": comment,
+            "submit": "Submit changes",
+        }
+        data.update(timestamps)
+        data.update(fields_data)
+
+        r = self.post("/ticket/{}".format(ticket_id), data)
+
+        # Starting from 1.0+, the system-message element is always on the page,
+        # only the style is changed.
+        if self.trac_version >= (1, 0):
+            token = 'system-message" style=""'
+        else:
+            token = "system-message"
+
+        if token in r.text or r.status_code != 200:
+            raise exceptions.FatalError("unable to save change")
+
     def run_comment(self, ticket_id):
         """Add a comment to the given ticket_id. This command does not return
         anything if successful. Command is cancelled if the content of the
@@ -457,10 +535,6 @@ class CartmanApp(object):
         """
         ticket_id = text.validate_id(ticket_id)
 
-        # Load the timestamps from the ticket page.
-        r = self.get("/ticket/{}".format(ticket_id))
-        timestamps = self._extract_timestamps(r.text)
-
         if self.message:
             comment = self.message
         else:
@@ -469,6 +543,12 @@ class CartmanApp(object):
         if not comment.strip():
             raise exceptions.FatalError("empty comment, cancelling")
 
+        self.login()
+
+        # Load the timestamps from the ticket page.
+        r = self.get("/ticket/{}".format(ticket_id))
+        timestamps = self._extract_timestamps(r.text)
+
         data = {
             "comment": comment,
             "action": "leave",
@@ -511,7 +591,14 @@ class CartmanApp(object):
         template = self.resolve_template()
 
         if not template:
-            template = DEFAULT_TEMPLATE
+            if self.message_file:
+                if self.message_file == "-":
+                    template = sys.stdin.read()
+                else:
+                    with open(self.message_file) as fp:
+                        template = fp.read()
+            else:
+                template = DEFAULT_TEMPLATE
 
         # Parse the template ahead of time, allowing us to insert the Owner/To.
         ep = email.parser.Parser()
@@ -527,11 +614,13 @@ class CartmanApp(object):
         if not headers["To"]:
             headers["To"] = self.username
 
+        self.login()
+
         valid = False
         while not valid:
             # Get the properties at each iteration, in case an admin updated
             # the list in the mean time.
-            properties = self.get_properties()
+            options = self.get_property_options()
 
             # Assume the user will produce a valid ticket
             valid = True
@@ -545,7 +634,10 @@ class CartmanApp(object):
             fp.write(body)
 
             fp.close()
-            self.editor(filename)
+
+            # When reading the message from stdin, we can't edit, skip editor.
+            if not self.message_file:
+                self.editor(filename)
 
             # Use the email parser to get the headers.
             ep = email.parser.Parser()
@@ -572,10 +664,10 @@ class CartmanApp(object):
             # try to complete them.
             for key in fuzzy_match_fields:
                 lkey = key.lower()
-                if lkey not in properties:
+                if lkey not in options:
                     continue
 
-                valid_options = properties[lkey]["options"]
+                valid_options = options[lkey]
 
                 # The specified value is not available in the multi-choice.
                 if key in headers and headers[key] not in valid_options:
@@ -602,11 +694,17 @@ class CartmanApp(object):
                     print(u" - {}".format(error))
 
                 try:
-                    self.input("\n-- Hit Enter to return to editor, "\
-                              "^C to abort --\n")
+                    if not self.message_file:
+                        self.input("\n-- Hit Enter to return to editor, "
+                                   "^C to abort --\n")
                 except KeyboardInterrupt:
                     raise exceptions.FatalError("ticket creation interrupted")
 
+            # There is no editor loop when reading message from stdin, just
+            # print the errors and exit.
+            if self.message_file:
+                break
+
         # Since the body is expected to be using CRLF line termination, we
         # replace newlines by CRLF if no CRLF is found.
         if "\r\n" not in body:
@@ -664,22 +762,21 @@ class CartmanApp(object):
         usage: cm properties
 
         """
-        properties = self.get_properties()
+        self.login()
 
-        return [
-            ui.title("Milestones"),
-            ", ".join(properties["milestone"]["options"]),
-            "",
-            ui.title("Components"),
-            ", ".join(properties["component"]["options"]),
-            "",
-            ui.title("Status"),
-            ", ".join(properties["status"]["options"]),
-            "",
-            ui.title("Priority"),
-            ", ".join(properties["priority"]["options"]),
-            "",
-        ]
+        options = self.get_property_options()
+
+        output = []
+        for title, prop in (("Milestones", "milestone"),
+                            ("Components", "component"),
+                            ("Status", "status"),
+                            ("Priorities", "priority")):
+            if prop in options:
+                output.extend([
+                    ui.title(title),
+                    ", ".join(options[prop]),
+                    ""])
+        return output
 
     def run_report(self, report_id=None):
         """List tickets from a given report number.
@@ -692,6 +789,8 @@ class CartmanApp(object):
 
         query_string = "/report/{}?format=tab".format(report_id)
 
+        self.login()
+
         for t in self.get_tickets(query_string):
             output.append(t.format_title())
 
@@ -705,6 +804,8 @@ class CartmanApp(object):
         """
         output = []
 
+        self.login()
+
         for d in self.get_dicts("/report?format=tab"):
             output.append("#{report}. {title}".format(**d))
 
@@ -721,12 +822,33 @@ class CartmanApp(object):
         output = []
         query_string = "/search?q={}".format("+".join(terms))
 
+        self.login()
+
         r = self.get(query_string)
         for ticket_id, description in text.extract_search_results(r.text):
             output.append("#{}. {}".format(ticket_id, description))
 
         return output
 
+    def run_timeline(self, daysback=None):
+        """Show tickets activity timeline
+
+        usage: cm timeline [daysback]
+
+        """
+        output = []
+        query_string = "/timeline?ticket=on"
+        if daysback:
+            query_string += "&daysback={}".format(daysback)
+
+        self.login()
+
+        r = self.get(query_string)
+        for item, description in text.extract_timeline_items(r.text):
+            output.append(u"{}. {}".format(item, description))
+
+        return output
+
     def run_status(self, ticket_id, status=None):
         """Updates the status of a ticket.
 
@@ -736,6 +858,8 @@ class CartmanApp(object):
         output = []
         ticket_id = text.validate_id(ticket_id)
 
+        self.login()
+
         # Get all the available actions for this ticket
         r = self.get("/ticket/{}".format(ticket_id))
         statuses = text.extract_statuses(r.text)
@@ -745,7 +869,8 @@ class CartmanApp(object):
             status = self.extract_status_from_ticket_page(r.text)
             output.append("Current status: {}".format(status))
             if statuses:
-                output.append("Available statuses: {}".format(", ".join(statuses)))
+                output.append("Available statuses: {}"
+                              .format(", ".join(statuses)))
             return output
 
         if not status:
@@ -779,6 +904,8 @@ class CartmanApp(object):
 
         query_string = "/ticket/{}?format=tab".format(ticket_id)
 
+        self.login()
+
         t = next(self.get_tickets(query_string))
         title = t.format_title()
 
