@@ -1,4 +1,4 @@
-# Copyright (c) 2011-2013 Bertrand Janin <b@janin.com>
+# Copyright (c) 2011-2015 Bertrand Janin <b@janin.com>
 #
 # Permission to use, copy, modify, and/or distribute this software for any
 # purpose with or without fee is hereby granted, provided that the above
@@ -28,6 +28,10 @@ re_search_result = re.compile(r'<dt><a href="[^"]+" class="searchable">'
                               r'<span class="\w+">#(\d+)</span>'
                               r': ([^<]+)</a></dt>')
 re_message = re.compile(r"<p class=\"message\">([^<]+)</p>")
+re_timeline_items = re.compile(
+                              r'<span class="time">[^<]+</span>[^<]*'
+                              r'<em[^>]*>([^<]*)</em>(.*)'
+                              )
 
 
 def fuzzy_find(value, options):
@@ -110,6 +114,7 @@ def extract_timestamps_v0(raw_html):
         "ts": extract_timestamps_common("ts", raw_html),
     }
 
+
 def extract_timestamps_v1(raw_html):
     return {
         "start_time": extract_timestamps_common("start_time", raw_html),
@@ -167,8 +172,8 @@ def extract_properties(raw_html):
     :param raw_html: Dump from the query page.
 
     """
-    re_prop = r"var properties=([^;]+)"
-    prop_tokens = re.findall(re_prop, raw_html, re.MULTILINE)
+    re_prop = r"var properties=(.+?);\s+var "
+    prop_tokens = re.findall(re_prop, raw_html, re.MULTILINE | re.DOTALL)
 
     if len(prop_tokens) < 1:
         return {}
@@ -217,3 +222,16 @@ def extract_search_results(raw_html):
     results = re_search_result.findall(raw_html)
 
     return [(int(r[0]), r[1]) for r in results]
+
+
+def extract_timeline_items(raw_html):
+    """Returns the timeline items.
+
+    TODO: include author
+
+    :param raw_html: Dump from any page.
+
+    """
+    results = re_timeline_items.findall(raw_html)
+
+    return [(r[0], r[1]) for r in results]
